import sys

import pytest

import fmmap as mmap


def test_off_by_one():
    with mmap.mmap(-1, 10) as m:
        m.write(b"jj________")
        m.seek(0)
        assert m.find(b"jj") == 0
        assert m.rfind(b"jj") == 0


def test_needle_too_big():
    with mmap.mmap(-1, 10) as m:
        m.write(b"jj________")
        m.seek(0)
        assert m.find(b"abc", 9) == -1
        assert m.rfind(b"abc", 9) == -1


def test_clean_namespace():
    # We shouldn't soil the module namespace with our own extras
    assert getattr(mmap, "version_info", None) is None
    assert getattr(mmap, "kernel", None) is None
    assert getattr(mmap, "OS", None) is None
    assert getattr(mmap, "uname", None) is None


@pytest.mark.skipif(not sys.platform.startswith("linux"), reason="Linux only")
def test_linux_constants():
    assert getattr(mmap, "MADV_DONTFORK") > 0
    assert getattr(mmap, "MADV_REMOVE") > 0


@pytest.mark.skipif(not sys.platform.startswith("freebsd"), reason="FreeBSD only")
def test_freebsd_constants():
    assert getattr(mmap, "MADV_FREE") > 0
    assert getattr(mmap, "MADV_NOSYNC") > 0


@pytest.mark.skipif(not sys.platform.startswith("openbsd"), reason="OpenBSD only")
def test_openbsd_constants():
    assert getattr(mmap, "MADV_FREE") > 0
    assert getattr(mmap, "MADV_SPACEAVAIL") > 0


@pytest.mark.skipif(not sys.platform.startswith("sunos"), reason="SunOS/Solaris only")
def test_openbsd_constants():
    assert getattr(mmap, "MADV_FREE") > 0
    assert getattr(mmap, "MADV_ACCESS_LWP") > 0



# From:
# https://foss.heptapod.net/pypy/pypy/raw/branch/py3.6/pypy/module/mmap/test/test_mmap.py
# Web view:
# https://foss.heptapod.net/pypy/pypy/blob/branch/py3.6/pypy/module/mmap/test/test_mmap.py

from __future__ import with_statement
#from rpython.tool.udir import udir
import os, sys, py
import shutil

import pytest
from pytest import raises, skip


import fmmap
sys.modules["mmap"] = fmmap

PLATFORMS_WITHOUT_MREMAP = (
    "darwin",
    "freebsd",
    "openbsd",
    "sunos",
)

class TestAppTestMMap:
    spaceconfig = dict(usemodules=('mmap',))


    def setup_class(cls):
        shutil.rmtree("./tmp", ignore_errors=True)
        os.mkdir('./tmp')
        cls.tmpname = "./tmp/mmap-"

    def teardown_class(cls):
        try:
            shutil.rmtree("./tmp")
        except FileNotFoundError:
            pass

    def test_page_size(self):
        import mmap
        assert mmap.PAGESIZE > 0
        assert mmap.ALLOCATIONGRANULARITY > 0
        assert isinstance(mmap.PAGESIZE, int)
        assert isinstance(mmap.ALLOCATIONGRANULARITY, int)
        assert mmap.ALLOCATIONGRANULARITY % mmap.PAGESIZE == 0

    def test_attributes(self):
        import mmap
        import os
        assert isinstance(mmap.ACCESS_READ, int)
        assert isinstance(mmap.ACCESS_WRITE, int)
        assert isinstance(mmap.ACCESS_COPY, int)
        if os.name == "posix":
            assert isinstance(mmap.MAP_ANON, int)
            assert isinstance(mmap.MAP_ANONYMOUS, int)
            assert isinstance(mmap.MAP_PRIVATE, int)
            assert isinstance(mmap.MAP_SHARED, int)
            assert isinstance(mmap.PROT_EXEC, int)
            assert isinstance(mmap.PROT_READ, int)
            assert isinstance(mmap.PROT_WRITE, int)

        assert mmap.error is OSError

    def test_args(self):
        from mmap import mmap
        import os
        import sys

        raises(TypeError, mmap, "foo")
        raises(TypeError, mmap, 0, "foo")

        if os.name == "posix":
            raises(ValueError, mmap, 0, 1, 2, 3, 4)
            raises(TypeError, mmap, 0, 1, 2, 3, "foo", 5)
            raises(TypeError, mmap, 0, 1, foo="foo")
            raises((TypeError, OverflowError), mmap, 0, -1)
            raises(OverflowError, mmap, 0, sys.maxsize ** 3)
            raises(ValueError, mmap, 0, 1, flags=2, access=3)
            raises(ValueError, mmap, 0, 1, access=123)
        elif os.name == "nt":
            raises(TypeError, mmap, 0, 1, 2, 3, 4)
            raises(TypeError, mmap, 0, 1, tagname=123)
            raises(TypeError, mmap, 0, 1, access="foo")
            raises(ValueError, mmap, 0, 1, access=-1)

    def test_subclass(self):
        import mmap
        class anon_mmap(mmap.mmap):
            def __new__(klass, *args, **kwargs):
                return mmap.mmap.__new__(klass, -1, *args, **kwargs)
        anon_mmap(mmap.PAGESIZE)

    def test_file_size(self):
        import os
        if os.name == "nt":
            skip("Only Unix checks file size")

        from mmap import mmap
        f = open(self.tmpname + "a", "wb+")

        f.write(b"c")
        f.flush()
        raises(ValueError, mmap, f.fileno(), 123)
        f.close()

    def test_create(self):
        from mmap import mmap
        f = open(self.tmpname + "b", "wb+")

        f.write(b"c")
        f.flush()
        m = mmap(f.fileno(), 1)
        assert m.read(99) == b"c"

        f.close()

    def test_close(self):
        from mmap import mmap
        f = open(self.tmpname + "c", "wb+")

        f.write(b"c")
        f.flush()
        m = mmap(f.fileno(), 1)
        m.close()
        raises(ValueError, m.read, 1)

    def test_read_byte(self):
        from mmap import mmap
        f = open(self.tmpname + "d", "wb+")

        f.write(b"c")
        f.flush()
        m = mmap(f.fileno(), 1)
        assert m.read_byte() == ord(b"c")
        raises(ValueError, m.read_byte)
        m.close()
        f.close()

    def test_readline(self):
        from mmap import mmap
        import os
        f = open(self.tmpname + "e", "wb+")

        f.write(b"foo\n")
        f.flush()
        m = mmap(f.fileno(), 4)
        assert m.readline() == b"foo\n"
        assert m.readline() == b""
        m.close()
        f.close()

    def test_read(self):
        from mmap import mmap
        f = open(self.tmpname + "f", "wb+")

        f.write(b"foobar")
        f.flush()
        m = mmap(f.fileno(), 6)
        raises(TypeError, m.read, b"foo")
        assert m.read(1) == b"f"
        assert m.read(6) == b"oobar"
        assert m.read(1) == b""
        m.close()
        f.close()

    def test_find(self):
        from mmap import mmap
        f = open(self.tmpname + "g", "wb+")

        f.write(b"foobar\0")
        f.flush()
        m = mmap(f.fileno(), 7)
        raises(TypeError, m.find, 123)
        raises(TypeError, m.find, b"foo", b"baz")
        assert m.find(b"b") == 3
        assert m.find(b"z") == -1
        assert m.find(b"o", 5) == -1
        assert m.find(b"ob") == 2
        assert m.find(b"\0") == 6
        assert m.find(b"ob", 1) == 2
        assert m.find(b"ob", 2) == 2
        assert m.find(b"ob", 3) == -1
        assert m.find(b"ob", -4) == -1
        assert m.find(b"ob", -5) == 2
        assert m.find(b"ob", -999999999) == 2
        assert m.find(b"ob", 1, 3) == -1
        assert m.find(b"ob", 1, 4) == 2
        assert m.find(b"ob", 1, 999999999) == 2
        assert m.find(b"ob", 1, 0) == -1
        assert m.find(b"ob", 1, -1) == 2
        assert m.find(b"ob", 1, -3) == 2
        assert m.find(b"ob", 1, -4) == -1
        #
        data = m.read(2)
        assert data == b"fo"
        assert m.find(b"o") == 2
        assert m.find(b"oo") == -1
        assert m.find(b"o", 0) == 1
        m.close()
        f.close()

    def test_rfind(self):
        from mmap import mmap
        f = open(self.tmpname + "g", "wb+")

        f.write(b"foobarfoobar\0")
        f.flush()
        m = mmap(f.fileno(), 13)
        raises(TypeError, m.rfind, 123)
        raises(TypeError, m.rfind, b"foo", b"baz")
        assert m.rfind(b"b") == 9
        assert m.rfind(b"z") == -1
        assert m.rfind(b"o", 11) == -1
        assert m.rfind(b"ob") == 8
        assert m.rfind(b"\0") == 12
        assert m.rfind(b"ob", 7) == 8
        assert m.rfind(b"ob", 8) == 8
        assert m.rfind(b"ob", 9) == -1
        assert m.rfind(b"ob", -4) == -1
        assert m.rfind(b"ob", -5) == 8
        assert m.rfind(b"ob", -999999999) == 8
        assert m.rfind(b"ob", 1, 3) == -1
        assert m.rfind(b"ob", 1, 4) == 2
        assert m.rfind(b"ob", 1, 999999999) == 8
        assert m.rfind(b"ob", 1, 0) == -1
        assert m.rfind(b"ob", 1, -1) == 8
        assert m.rfind(b"ob", 1, -3) == 8
        assert m.rfind(b"ob", 1, -4) == 2
        #
        data = m.read(8)
        assert data == b"foobarfo"
        assert m.rfind(b"o") == 8
        assert m.rfind(b"oo") == -1
        assert m.rfind(b"o", 0) == 8
        m.close()
        f.close()

    def test_is_modifiable(self):
        import mmap
        f = open(self.tmpname + "h", "wb+")

        f.write(b"foobar")
        f.flush()
        m = mmap.mmap(f.fileno(), 6, access=mmap.ACCESS_READ)
        raises(TypeError, m.write, b'x')
        raises(TypeError, m.resize, 7)
        m.close()
        f.close()

    def test_seek(self):
        from mmap import mmap
        f = open(self.tmpname + "i", "wb+")

        f.write(b"foobar")
        f.flush()
        m = mmap(f.fileno(), 6)
        raises(TypeError, m.seek, b"foo")
        raises(TypeError, m.seek, 0, b"foo")
        raises(ValueError, m.seek, -1, 0)
        raises(ValueError, m.seek, -1, 1)
        raises(ValueError, m.seek, -7, 2)
        raises(ValueError, m.seek, 1, 3)
        raises(ValueError, m.seek, 10)
        m.seek(0)
        assert m.tell() == 0
        m.read(1)
        m.seek(1, 1)
        assert m.tell() == 2
        m.seek(0)
        m.seek(-1, 2)
        assert m.tell() == 5
        m.close()
        f.close()

    def test_write(self):
        import mmap
        f = open(self.tmpname + "j", "wb+")

        f.write(b"foobar")
        f.flush()
        m = mmap.mmap(f.fileno(), 6, access=mmap.ACCESS_READ)
        raises(TypeError, m.write, b"foo")
        m = mmap.mmap(f.fileno(), 6, access=mmap.ACCESS_WRITE)
        raises(TypeError, m.write, 123)
        raises(ValueError, m.write, b"c"*10)
        assert m.write(b"ciao\n") == 5
        m.seek(0)
        assert m.read(6) == b"ciao\nr"
        m.close()

    def test_write_byte(self):
        import mmap
        f = open(self.tmpname + "k", "wb+")

        f.write(b"foobar")
        f.flush()
        m = mmap.mmap(f.fileno(), 6, access=mmap.ACCESS_READ)
        raises(TypeError, m.write_byte, b"f")
        m = mmap.mmap(f.fileno(), 6, access=mmap.ACCESS_WRITE)
        raises(TypeError, m.write_byte, "a")
        raises(TypeError, m.write_byte, b"a")
        m.write_byte(ord("x"))
        m.seek(0)
        assert m.read(6) == b"xoobar"
        m.close()

    def test_size(self):
        from mmap import mmap
        f = open(self.tmpname + "l", "wb+")

        f.write(b"foobar")
        f.flush()
        m = mmap(f.fileno(), 5)
        assert m.size() == 6 # size of the underline file, not the mmap
        m.close()
        f.close()

    def test_tell(self):
        from mmap import mmap
        f = open(self.tmpname + "m", "wb+")

        f.write(b"c")
        f.flush()
        m = mmap(f.fileno(), 1)
        assert m.tell() >= 0
        m.close()
        f.close()

    def test_flush(self):
        from mmap import mmap
        f = open(self.tmpname + "n", "wb+")

        f.write(b"foobar")
        f.flush()
        m = mmap(f.fileno(), 6)
        raises(TypeError, m.flush, 1, 2, 3)
        raises(TypeError, m.flush, 1, b"a")
        raises(ValueError, m.flush, 0, 99)
        m.flush()    # return value is a bit meaningless, platform-dependent
        m.close()
        f.close()

    def test_length_0_large_offset(self):
        import mmap

        with open(self.tmpname, "wb") as f:
            f.write(115699 * b'm')
        with open(self.tmpname, "w+b") as f:
            raises(ValueError, mmap.mmap, f.fileno(), 0, offset=2147418112)

    def test_move(self):
        import mmap
        f = open(self.tmpname + "o", "wb+")

        f.write(b"foobar")
        f.flush()
        m = mmap.mmap(f.fileno(), 6, access=mmap.ACCESS_READ)
        raises(TypeError, m.move, 1)
        raises(TypeError, m.move, 1, b"foo", 2)
        m = mmap.mmap(f.fileno(), 6, access=mmap.ACCESS_WRITE)
        raises(ValueError, m.move, 7, 1, 2)
        raises(ValueError, m.move, 1, 7, 2)
        m.move(1, 3, 3)
        assert m.read(6) == b"fbarar"
        m.seek(0)
        m.move(1, 3, 2)
        a = m.read(6)
        assert a == b"frarar"
        m.close()
        f.close()

    def test_resize(self):
        import sys
        for platform in PLATFORMS_WITHOUT_MREMAP:
            if platform in sys.platform:
                skip("resize does not work under some platforms")

        import mmap
        import os

        f = open(self.tmpname + "p", "wb+")
        f.write(b"foobar")
        f.flush()
        m = mmap.mmap(f.fileno(), 6, access=mmap.ACCESS_READ)
        raises(TypeError, m.resize, 1)
        m = mmap.mmap(f.fileno(), 6, access=mmap.ACCESS_COPY)
        raises(TypeError, m.resize, 1)
        m = mmap.mmap(f.fileno(), 6, access=mmap.ACCESS_WRITE)
        f_size = os.fstat(f.fileno()).st_size
        assert m.size() == f_size == 6
        m.resize(10)
        f_size = os.fstat(f.fileno()).st_size
        assert m.size() == f_size == 10
        m.close()
        f.close()

    def test_resize_bsd(self):
        import sys
        for platform in PLATFORMS_WITHOUT_MREMAP:
            if platform in sys.platform:
                break
        else:
            skip("test different resize on some platforms")

        import mmap
        import os

        f = open(self.tmpname + "p", "wb+")
        f.write(b"foobar")
        f.flush()
        m = mmap.mmap(f.fileno(), 6, access=mmap.ACCESS_READ)
        raises(TypeError, m.resize, 1)
        m = mmap.mmap(f.fileno(), 6, access=mmap.ACCESS_COPY)
        raises(TypeError, m.resize, 1)
        m = mmap.mmap(f.fileno(), 6, access=mmap.ACCESS_WRITE)
        f_size = os.fstat(f.fileno()).st_size
        assert m.size() == f_size == 6
        raises(SystemError, m.resize, 10)
        f_size = os.fstat(f.fileno()).st_size
        assert m.size() == f_size == 6

    def test_len(self):
        from mmap import mmap

        f = open(self.tmpname + "q", "wb+")
        f.write(b"foobar")
        f.flush()

        m = mmap(f.fileno(), 6)
        assert len(m) == 6
        m.close()
        f.close()

    def test_get_item(self):
        from mmap import mmap

        f = open(self.tmpname + "r", "wb+")
        f.write(b"foobar")
        f.flush()

        m = mmap(f.fileno(), 6)
        fn = lambda: m[b"foo"]
        raises(TypeError, fn)
        fn = lambda: m[-7]
        raises(IndexError, fn)
        assert m[0] == ord('f')
        assert m[-1] == ord('r')
        assert m[1::2] == b'obr'
        assert m[4:1:-2] == b'ao'
        m.close()
        f.close()

    def test_get_crash(self):
        import sys
        from mmap import mmap
        s = b'hallo!!!'
        m = mmap(-1, len(s))
        m[:] = s
        assert m[1:None:sys.maxsize] == b'a'
        m.close()

    def test_set_item(self):
        import mmap

        f = open(self.tmpname + "s", "wb+")
        f.write(b"foobar")
        f.flush()

        m = mmap.mmap(f.fileno(), 6, access=mmap.ACCESS_READ)
        def fn(): m[1] = b'a'
        raises(TypeError, fn)
        m = mmap.mmap(f.fileno(), 6, access=mmap.ACCESS_WRITE)
        def fn(): m[b"foo"] = b'a'
        raises(TypeError, fn)
        def fn(): m[-7] = b'a'
        raises(IndexError, fn)
        def fn(): m[1:3] = 'xx'
        raises((IndexError, TypeError), fn)      # IndexError is in CPython,
                                                 # but doesn't make much sense
        def fn(): m[1:4] = b"zz"
        raises((IndexError, ValueError), fn)
        def fn(): m[1:6] = b"z" * 6
        raises((IndexError, ValueError), fn)
        def fn(): m[:2] = b"z" * 5
        raises((IndexError, ValueError), fn)
        def fn(): m[0] = 256
        raises(ValueError, fn)
        m[1:3] = b'xx'
        assert m.read(6) == b"fxxbar"
        m[0] = ord('x')
        assert m[0] == ord('x')
        m[-6] = ord('y')
        m[3:6:2] = b'BR'
        m.seek(0)
        data = m.read(6)
        assert data == b"yxxBaR"
        m.close()
        f.close()

    def test_del_item(self):
        from mmap import mmap

        f = open(self.tmpname + "t", "wb+")
        f.write(b"foobar")
        f.flush()

        m = mmap(f.fileno(), 6)
        def fn(): del m[b"foo"]
        raises(TypeError, fn)
        def fn(): del m[1:3]
        raises(TypeError, fn)
        def fn(): del m[1]
        raises(TypeError, fn)
        m.close()
        f.close()

    def test_concatenation(self):
        from mmap import mmap

        f = open(self.tmpname + "u", "wb+")
        f.write(b"foobar")
        f.flush()

        m = mmap(f.fileno(), 6)
        def fn(): m + 1
        raises((SystemError, TypeError), fn)     # SystemError is in CPython,
        def fn(m): m += 1                        # but it doesn't make much
        raises((SystemError, TypeError), fn, m)  # sense
        def fn(): 1 + m
        raises(TypeError, fn)
        m.close()
        f.close()

    def test_repeatition(self):
        from mmap import mmap

        f = open(self.tmpname + "v", "wb+")
        f.write(b"foobar")
        f.flush()

        m = mmap(f.fileno(), 6)
        def fn(): m * 1
        raises((SystemError, TypeError), fn)      # SystemError is in CPython,
        def fn(m): m *= 1                         # but it
        raises((SystemError, TypeError), fn, m)   # doesn't
        def fn(): 1 * m                           # make much sense
        raises((SystemError, TypeError), fn)
        m.close()
        f.close()

    def test_slicing(self):
        from mmap import mmap

        f = open(self.tmpname + "v", "wb+")
        f.write(b"foobar")
        f.flush()

        f.seek(0)
        m = mmap(f.fileno(), 6)
        assert m[-3:7] == b"bar"

        assert m[1:0:1] == b""

        f.close()

    def test_memoryview(self):
        from mmap import mmap, ACCESS_READ
        filename = self.tmpname + "y"
        f = open(filename, "bw+")
        f.write(b"foobar")
        f.flush()
        m = mmap(f.fileno(), 6)
        b = memoryview(m)
        assert len(b) == 6
        assert b.readonly is False
        assert b[3] == ord(b"b")
        assert b[:] == b"foobar"
        del b  # For CPython: "exported pointers exist"
        m.close()
        f.close()
        with open(filename, "rb") as f:
            m = mmap(f.fileno(), 6, access=ACCESS_READ)
            b = memoryview(m)
            assert b.readonly is True
            assert b[:] == b"foobar"
            del b
            m.close()

    def test_offset(self):
        from mmap import mmap, ALLOCATIONGRANULARITY
        filename = self.tmpname + "y"
        f = open(filename, "wb+")
        f.write(b"foobar" * ALLOCATIONGRANULARITY)
        f.flush()
        size = ALLOCATIONGRANULARITY
        offset = 2 * ALLOCATIONGRANULARITY
        m = mmap(f.fileno(), size, offset=offset)
        assert m[:] == (b"foobar" * ALLOCATIONGRANULARITY)[offset:offset+size]
        assert len(m) == size
        m.close()
        f.close()

    def test_offset_more(self):
        from mmap import mmap, ALLOCATIONGRANULARITY

        with open(self.tmpname, "w+b") as f:
            halfsize = ALLOCATIONGRANULARITY
            f.write(b"\0" * halfsize)
            f.write(b"foo")
            f.write(b"\0" * (halfsize - 3))
            f.flush()
            m = mmap(f.fileno(), 0)
            m.close()

        with open(self.tmpname, "r+b") as f:
            m = mmap(f.fileno(), halfsize, offset=halfsize)
            assert m[0:3] == b"foo"

        try:
            m.resize(512)
        except SystemError:
            pass
        else:
            assert len(m) == 512
            raises(ValueError, m.seek, 513, 0)
            assert m[0:3] == b"foo"
            with open(self.tmpname) as f:
                f.seek(0, 2)
                assert f.tell() == halfsize + 512
            assert m.size() == halfsize + 512
        m.close()

    def test_large_offset(self):
        import mmap
        import sys
        size = 0x14FFFFFFF
        if sys.platform.startswith('win') or sys.platform == 'darwin':
            skip('test requires %s bytes and a long time to run' % size)

        with open(self.tmpname, "w+b") as f:
            f.seek(size)
            f.write(b"A")
            f.flush()
        with open(self.tmpname, 'rb') as f2:
            f2.seek(size)
            c = f2.read(1)
            assert c == b'A'
            m = mmap.mmap(f2.fileno(), 0, offset=0x140000000,
                          access=mmap.ACCESS_READ)
            try:
                assert m[0xFFFFFFF] == ord('A')
            finally:
                m.close()
    test_large_offset.is_large = True

    def test_large_filesize(self):
        import mmap
        import sys
        size = 0x17FFFFFFF
        if sys.platform.startswith('win') or sys.platform == 'darwin':
            skip('test requires %s bytes and a long time to run' % size)

        with open(self.tmpname, "w+b") as f:
            f.seek(size)
            f.write(b" ")
            f.flush()
            m = mmap.mmap(f.fileno(), 0x10000, access=mmap.ACCESS_READ)
            try:
                assert m.size() ==  0x180000000
            finally:
                m.close()
    test_large_filesize.is_large = True

    def test_context_manager(self):
        import mmap
        with mmap.mmap(-1, 10) as m:
            assert not m.closed
        assert m.closed

    def test_all(self):
        # this is a global test, ported from test_mmap.py
        import mmap
        from mmap import PAGESIZE
        import sys
        import os

        filename = self.tmpname + "w"

        f = open(filename, "wb+")

        # write 2 pages worth of data to the file
        f.write(b'\0' * PAGESIZE)
        f.write(b'foo')
        f.write(b'\0' * (PAGESIZE - 3))
        f.flush()
        m = mmap.mmap(f.fileno(), 2 * PAGESIZE)
        f.close()

        # sanity checks
        assert m.find(b"foo") == PAGESIZE
        assert len(m) == 2 * PAGESIZE
        assert m[0] == 0
        assert m[0:3] == b'\0\0\0'

        # modify the file's content
        m[0] = ord('3')
        m[PAGESIZE+3:PAGESIZE+3+3] = b'bar'

        # check that the modification worked
        assert m[0] == ord('3')
        assert m[0:3] == b'3\0\0'
        assert m[PAGESIZE-1:PAGESIZE+7] == b'\0foobar\0'

        m.flush()

        # test seeking around
        m.seek(0,0)
        assert m.tell() == 0
        m.seek(42, 1)
        assert m.tell() == 42
        m.seek(0, 2)
        assert m.tell() == len(m)

        raises(ValueError, m.seek, -1)
        raises(ValueError, m.seek, 1, 2)
        raises(ValueError, m.seek, -len(m) - 1, 2)

        # try resizing map
        for platform in PLATFORMS_WITHOUT_MREMAP:
            if platform in sys.platform:
                break
        else:
            m.resize(512)

            assert len(m) == 512
            raises(ValueError, m.seek, 513, 0)

            # check that the underlying file is truncated too
            f = open(filename)
            f.seek(0, 2)
            assert f.tell() == 512
            f.close()
            assert m.size() == 512

        m.close()
        f.close()

        # test access=ACCESS_READ
        mapsize = 10
        f = open(filename, "wb")
        f.write(b"a" * mapsize)
        f.close()
        f = open(filename, "rb")
        m = mmap.mmap(f.fileno(), mapsize, access=mmap.ACCESS_READ)
        assert m[:] == b'a' * mapsize
        def f(m): m[:] = b'b' * mapsize
        raises(TypeError, f, m)
        def fn(): m[0] = b'b'
        raises(TypeError, fn)
        def fn(m): m.seek(0, 0); m.write(b"abc")
        raises(TypeError, fn, m)
        def fn(m): m.seek(0, 0); m.write_byte(b"d")
        raises(TypeError, fn, m)
        if not (("darwin" in sys.platform) or ("freebsd" in sys.platform)):
            raises(TypeError, m.resize, 2 * mapsize)
            assert open(filename, "rb").read() == b'a' * mapsize

        # opening with size too big
        f = open(filename, "r+b")
        if not os.name == "nt":
            # this should work under windows
            raises(ValueError, mmap.mmap, f.fileno(), mapsize + 1)
        f.close()

        # if _MS_WINDOWS:
        #     # repair damage from the resizing test.
        #     f = open(filename, 'r+b')
        #     f.truncate(mapsize)
        #     f.close()
        m.close()

        # test access=ACCESS_WRITE"
        f = open(filename, "r+b")
        m = mmap.mmap(f.fileno(), mapsize, access=mmap.ACCESS_WRITE)
        m.write(b'c' * mapsize)
        m.seek(0)
        data = m.read(mapsize)
        assert data == b'c' * mapsize
        m.flush()
        m.close()
        f.close()
        f = open(filename, 'rb')
        stuff = f.read()
        f.close()
        assert stuff == b'c' * mapsize

        # test access=ACCESS_COPY
        f = open(filename, "r+b")
        m = mmap.mmap(f.fileno(), mapsize, access=mmap.ACCESS_COPY)
        m.write(b'd' * mapsize)
        m.seek(0)
        data = m.read(mapsize)
        assert data == b'd' * mapsize
        m.flush()
        assert open(filename, "rb").read() == b'c' * mapsize
        if not (("darwin" in sys.platform) or ("freebsd" in sys.platform)):
            raises(TypeError, m.resize, 2 * mapsize)
        m.close()
        f.close()

        # test invalid access
        f = open(filename, "r+b")
        raises(ValueError, mmap.mmap, f.fileno(), mapsize, access=4)
        f.close()

        # test incompatible parameters
        if os.name == "posix":
            f = open(filename, "r+b")
            raises(ValueError, mmap.mmap, f.fileno(), mapsize, flags=mmap.MAP_PRIVATE,
                prot=mmap.PROT_READ, access=mmap.ACCESS_WRITE)
            f.close()


        # bad file descriptor
        raises(EnvironmentError, mmap.mmap, -2, 4096)

        # do a tougher .find() test.  SF bug 515943 pointed out that, in 2.2,
        # searching for data with embedded \0 bytes didn't work.
        f = open(filename, 'wb+')
        data = b'aabaac\x00deef\x00\x00aa\x00'
        n = len(data)
        f.write(data)
        f.flush()
        m = mmap.mmap(f.fileno(), n)
        f.close()

        for start in range(n + 1):
            for finish in range(start, n + 1):
                sl = data[start:finish]
                assert m.find(sl) == data.find(sl)
                assert m.find(sl + b'x') ==  -1
        m.close()

        # test mapping of entire file by passing 0 for map length
        f = open(filename, "wb+")
        f.write(2**16 * b'm')
        f.close()
        f = open(filename, "rb+")
        m = mmap.mmap(f.fileno(), 0)
        assert len(m) == 2**16
        assert m.read(2**16) == 2**16 * b"m"
        m.close()
        f.close()

        # make move works everywhere (64-bit format problem earlier)
        f = open(filename, 'wb+')
        f.write(b"ABCDEabcde")
        f.flush()
        m = mmap.mmap(f.fileno(), 10)
        m.move(5, 0, 5)
        assert m.read(10) == b"ABCDEABCDE"
        m.close()
        f.close()

    def test_empty_file(self):
        import mmap
        f = open(self.tmpname, 'w+b')
        f.close()
        with open(self.tmpname, 'rb') as f:
            try:
                m = mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)
                m.close()
                assert False, "should not have been able to mmap empty file"
            except ValueError as e:
                assert str(e) == "cannot mmap an empty file"
            except BaseException as e:
                assert False, "unexpected exception: " + str(e)

    def test_read_all(self):
        from mmap import mmap
        f = open(self.tmpname + "f", "wb+")
        f.write(b"foobar")
        f.flush()

        m = mmap(f.fileno(), 6)
        assert m.read(None) == b"foobar"

    def test_resize_past_pos(self):
        import os, mmap, sys
        if os.name == "nt":
            skip("cannot resize anonymous mmaps on Windows")
        if sys.version_info < (2, 7, 13):
            skip("cannot resize anonymous mmaps before 2.7.13")
        m = mmap.mmap(-1, 8192)
        m.read(5000)
        try:
            m.resize(4096)
        except SystemError:
            skip("resizing not supported")
        assert m.tell() == 5000
        assert m.read(14) == b''
        assert m.read(-1) == b''
        raises(ValueError, m.read_byte)
        assert m.readline() == b''
        raises(ValueError, m.write_byte, ord(b'b'))
        raises(ValueError, m.write, b'abc')
        assert m.tell() == 5000
        m.close()




#!/usr/bin/env python

import mmap
import sys
import timeit

import fmmap


NUMBER=5


needle = sys.argv[1].encode('utf-8')


def find(m, needle=needle):
    return m.find(needle)


def rfind(m, needle=needle):
    return m.rfind(needle)


for name in sys.argv[2:]:
    with open(name, 'rb') as f:

        # built-in mmap
        with mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ) as m:
            a_find = find(m)
            print("mmap: find:", a_find)
            a1 = timeit.timeit('find(m)', number=NUMBER, globals={"m": m, "find": find})
            print(f"{a1:.3f}")
            a_rfind = rfind(m)
            print("mmap: rfind:", a_rfind)
            a2 = timeit.timeit('find(m)', number=NUMBER, globals={"m": m, "find": rfind})
            print(f"{a2:.3f}")

        # fmmap
        with fmmap.mmap(f.fileno(), 0, access=fmmap.ACCESS_READ) as m:
            b_find = find(m)
            print("fmmap: find:", a_find)
            b1 = timeit.timeit('find(m)', number=NUMBER, globals={"m": m, "find": find})
            print(f"{b1:.3f}")
            b_rfind = rfind(m)
            print("fmmap: rfind:", b_rfind)
            b2 = timeit.timeit('find(m)', number=NUMBER, globals={"m": m, "find": rfind})
            print(f"{b2:.3f}")

        if a_find != b_find:
            print("Results for find() differs!")
        if a_rfind != b_rfind:
            print("Results for rfind() differs!")

        print()
        print("Ratio mmap:fmmap (bigger than 1.0 means fmmap is faster)")
        print(f"find:  {a1/b1:2.2f}")
        print(f"rfind: {a2/b2:2.2f}")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# cython: language_level=3

from mmap import *
import sys


cdef str platform = <str>sys.platform

cdef int py_version = sys.hexversion
DEF PY39 = 0x030901f0
DEF PY38 = 0x030801f0
DEF PY37 = 0x030701f0
DEF PY36 = 0x030601f0


import cython
from cpython cimport exc
from libc cimport string

cimport constants


cdef extern from "<string.h>":
    #GNU extension to glibc
    unsigned char *memmem(const void *haystack, size_t haystacklen,
                 const void *needle, size_t needlelen) nogil


# See man memmem(3)
cdef unsigned char *my_memmem(
        const unsigned char *buf_p,
        size_t haystack_len,
        const unsigned char *needle,
        size_t needle_len,
    ) nogil:
    cdef unsigned char *c
    i = haystack_len - needle_len + 1
    c = <unsigned char *>string.memchr(buf_p, needle[0], i)
    while c:
        if string.memcmp(c, needle, needle_len) == 0:
            return c
        i = haystack_len - (c - buf_p) - needle_len
        c = <unsigned char *>string.memchr(c + 1, needle[0], i)

    return NULL


# See man memchr(3)
cdef unsigned char *my_memrchr(
        unsigned char *b,
        int c,
        size_t n,
    ) nogil:
    b += n - 1
    while n:
        if <unsigned char>c == b[0]:
            return b
        b -= 1
        n -= 1

    return NULL


_transform_flush_return_value = lambda value: value


if py_version < PY38:
    # We want to implement the return convention for flush() introduced in
    # Python 3.8. If we are running on an earlier version, let's massage the
    # return value:

    IF UNAME_SYSNAME == "Windows":
        def _transform_flush_return_value(value):
            if value == 0:
                # error
                exc.PyErr_SetFromWindowsErr(0)
            else:
                # success
                return None
    ELSE:
        def _transform_flush_return_value(value):
            # unix (and others)
            if value == 0:
                # success
                return None
            else:
                # error
                # Should not be reached, since flush() raises an exception on
                # errors on Python < 3.8
                pass


IF UNAME_SYSNAME != "Windows":
    if py_version < PY38:
        # Constants needed for madvise

        from posix cimport mman

        MADV_NORMAL = mman.MADV_NORMAL
        MADV_RANDOM = mman.MADV_RANDOM
        MADV_SEQUENTIAL = mman.MADV_SEQUENTIAL
        MADV_WILLNEED = mman.MADV_WILLNEED
        MADV_DONTNEED = mman.MADV_DONTNEED
        # common in several Unix type systems;
        if constants.FREE:
            MADV_FREE = mman.MADV_FREE

        IF UNAME_SYSNAME == "Linux":

            from platform import uname

            kernel = tuple(int(x) for x in uname()[2].split('-')[0].split('.'))
            if kernel >= (2, 6, 16):
                MADV_REMOVE = mman.MADV_REMOVE
                MADV_DONTFORK = mman.MADV_DONTFORK
                MADV_DOFORK = mman.MADV_DOFORK
            if kernel >= (2, 6, 32):
                if constants.HWPOISON:
                    MADV_HWPOISON = mman.MADV_HWPOISON
                if constants.MERGEABLE:
                    MADV_MERGEABLE = mman.MADV_MERGEABLE
                    MADV_UNMERGEABLE = mman.MADV_UNMERGEABLE
            if kernel >= (2, 6, 33) and constants.SOFT_OFFLINE:
                MADV_SOFT_OFFLINE = mman.MADV_SOFT_OFFLINE
            if kernel >= (2, 6, 38) and constants.HUGEPAGE:
                MADV_HUGEPAGE = mman.MADV_HUGEPAGE
                MADV_NOHUGEPAGE = mman.MADV_NOHUGEPAGE
            if kernel >= (3, 4, 0) and constants.DUMP:
                MADV_DONTDUMP = mman.MADV_DONTDUMP
                MADV_DODUMP = mman.MADV_DODUMP
            if kernel >= (4, 14, 0) and constants.ONFORK:
                MADV_WIPEONFORK = mman.MADV_WIPEONFORK
                MADV_KEEPONFORK = mman.MADV_KEEPONFORK
            del kernel
            del uname

        ELSE:
            # FreeBSD:
            if constants.NOSYNC:
                MADV_NOSYNC = constants.MADV_NOSYNC
                MADV_AUTOSYNC = constants.MADV_AUTOSYNC
            if constants.CORE:
                MADV_NOCORE = constants.MADV_NOCORE
                MADV_CORE = constants.MADV_CORE
            if constants.PROTECT:
                MADV_PROTECT = constants.MADV_PROTECT


    # Some madvise constants aren't in the standard library (in any Python
    # version so far), so we expose them here unconditionally if they are in
    # <sys/mman.h>.

    # OpenBSD:
    if constants.SPACEAVAIL:
        MADV_SPACEAVAIL = constants.MADV_SPACEAVAIL

    # Solaris/illumos/OpenIndiana/SmartOs:
    if constants.ACCESS_DEFAULT:
        MADV_ACCESS_DEFAULT = constants.MADV_ACCESS_DEFAULT
    if constants.ACCESS_LWP:
        MADV_ACCESS_LWP = constants.MADV_ACCESS_LWP
    if constants.ACCESS_MANY:
        MADV_ACCESS_MANY = constants.MADV_ACCESS_MANY
    if constants.ACCESS_MANY_PSET:
        MADV_ACCESS_MANY_PSET = constants.MADV_ACCESS_MANY_PSET
    #illumos/OpenIndiana/SmartOS:
    if constants.PURGE:
        MADV_PURGE = constants.MADV_PURGE


if py_version < PY37:
    ACCESS_DEFAULT = 0


_mmap = mmap

class mmap(_mmap):

    if py_version < PY39:

        def __init__(self, *args, **kwargs):
            self._fileno = kwargs.get("fileno", args[0])
            # remember a few parameters for __repr__
            self._access = kwargs.get("access", 0)  # kwarg only
            self._offset = kwargs.get("offset", 0)  # kwarg only
            _mmap.__init__(*args, **kwargs)

        def __repr__(self):
            if self.closed:
                return "<fmmap.mmap closed=True>"
            names = {
                    ACCESS_DEFAULT: "ACCESS_DEFAULT",
                    ACCESS_READ: "ACCESS_READ",
                    ACCESS_WRITE: "ACCESS_WRITE",
                    ACCESS_COPY: "ACCESS_COPY",
            }
            access = names.get(self._access, "unknown!")
            return ("<fmmap.mmap "
                f"closed=False, "
                f"access={access}, "
                f"length={len(self)}, "
                f"pos={self.tell()}, "
                f"offset={self._offset}>"
            )

    IF UNAME_SYSNAME != "Windows":
        if py_version < PY38:

            def madvise(self, option, start=0, length=None):
                cdef const unsigned char[:] buf = self
                cdef ssize_t buf_len = len(buf)
                cdef unsigned char *buf_p

                if length is None:
                    length = buf_len

                if start < 0 or start >= buf_len:
                    raise ValueError("madvise start out of bounds")
                if length < 0:
                    raise ValueError("madvise length invalid")
                if sys.maxsize - start < length:
                    raise OverflowError("madvise length too large")

                if start + length > buf_len:
                    length = buf_len - start

                buf_p = &buf[start]
                if mman.madvise(buf_p, length, option) != 0:
                    exc.PyErr_SetFromErrno(OSError)

    if py_version < PY38:

        def flush(self, *args, **kwargs):
            value = super().flush(*args, **kwargs)
            return _transform_flush_return_value(value)

    if py_version < PY36:

        def __add__(self, value):
            raise TypeError()

        def __mul__(self, value):
            raise TypeError()

        def write(self, bytes):
            cdef const unsigned char[:] buf = bytes
            super().write(buf)
            return len(buf)

        def resize(self, newsize):
            if self._access not in (ACCESS_WRITE, ACCESS_DEFAULT):
                raise TypeError()
            if newsize < 0 or sys.maxsize - newsize < self._offset:
                raise ValueError("new size out of range")
            if self._fileno != -1:
                super().resize(newsize)
                return

            # There is a bug in Python versions before 3.6. It would call
            # ftruncate(2) on file descriptor -1 (anonymous memory), so we
            # can't fall back on the built-in implementation.
            raise SystemError("Can't resize anonymous memory in Python < 3.6")


    def find(object self, sub, start=None, end=None):
        cdef const unsigned char[:] buf = self
        if start is None:
            start = self.tell()
        if end is None:
            end = len(buf)
        return self._find(sub, start, end)

    @cython.boundscheck(False)
    def _find(object self, r, ssize_t start, ssize_t end):
        cdef const unsigned char[:] buf = self
        cdef const unsigned char[:] needle = r
        cdef ssize_t buf_len = len(buf)
        cdef ssize_t needle_len = len(needle)
        cdef unsigned char *c
        cdef unsigned char *buf_p
        cdef unsigned char *needle_p

        # negative slicing and bounds checking
        if start < 0:
            start += buf_len
            if start < 0:
                start = 0
        elif start > buf_len:
            start = buf_len
        if end < 0:
            end += buf_len
            if end < 0:
                end = 0
        elif end > buf_len:
            end = buf_len

        # trivial cases
        if start >= end:
            return -1
        if needle_len == 0:
            return 0
        if buf_len == 0 or needle_len > buf_len:
            return -1
        if end - start < needle_len:
            return -1

        with nogil:
            buf_p = &buf[start]
            needle_p = &needle[0]
            if constants.MEMMEM:
                c = memmem(buf_p, end-start, needle_p, needle_len)
            else:
                c = my_memmem(buf_p, end-start, needle_p, needle_len)

        if c is NULL:
            return -1
        return c - buf_p + start
        
    def rfind(object self, sub, start=None, end=None):
        cdef const unsigned char[:] buf = self
        if start is None:
            start = self.tell()
        if end is None:
            end = len(buf)
        return self._rfind(sub, start, end)

    @cython.boundscheck(False)
    def _rfind(object self, r, ssize_t start, ssize_t end):
        cdef const unsigned char[:] buf = self
        cdef const unsigned char[:] needle = r
        cdef ssize_t buf_len = len(buf)
        cdef ssize_t needle_len = len(needle)
        cdef unsigned char *c = NULL
        cdef unsigned char *buf_p
        cdef unsigned char *needle_p
        cdef ssize_t i

        # negative slicing and bounds checking
        if start < 0:
            start += buf_len
            if start < 0:
                start = 0
        elif start > buf_len:
            start = buf_len
        if end < 0:
            end += buf_len
            if end < 0:
                end = 0
        elif end > buf_len:
            end = buf_len

        # trivial cases
        if start >= end:
            return -1
        if needle_len == 0:
            return 0
        if buf_len == 0 or needle_len > buf_len:
            return -1
        if needle_len > end - start:
            return -1

        with nogil:
            # Maybe not as fast as a good memmem(), but memrchr is hopefully
            # optimised. Worst case is still O(nm) where
            #  - n = len(buf)
            #  - m = len(needle)
            # Hopefully it is still faster than the naive algorithm in
            # CPython, or looping ourselves here.
            #
            # We repeatedly search for the first byte of needle from end to
            # start. When finding it, we check if the whole needle is there.

            buf_p = &buf[start]
            needle_p = &needle[0]
            i = end - start - needle_len + 1
            if constants.MEMRCHR:
                c = <unsigned char *>string.memrchr(buf_p, needle[0], i)
                while c:
                    if string.memcmp(c, needle_p, needle_len) == 0:
                        break
                    c = <unsigned char *>string.memrchr(buf_p, needle[0], c - buf_p)

            else:
                c = my_memrchr(buf_p, needle[0], i)
                while c:
                    if string.memcmp(c, needle_p, needle_len) == 0:
                        break
                    c = my_memrchr(buf_p, needle[0], c - buf_p)

        if c is NULL:
            return -1
        return c - buf_p + start



# Some of the constants for madvise(2) are not guaranteed to be available.
# Apart from kernel version, it might be configured out, and simply not defined
# in the headers. While Cython lacks a #ifndef equivalent, we have to define a
# few things in a header file so that the names are defined so that the
# generated C file will compile.

cdef extern from "constants.h":

    # feature availability flags:
    enum: MEMMEM
    enum: MEMRCHR


    # constants related to madvise(2):

    # common
    enum: FREE

    # Linux
    enum: HWPOISON
    enum: MERGEABLE
    enum: SOFT_OFFLINE
    enum: HUGEPAGE
    enum: DUMP
    enum: ONFORK

    # In the following cases, the MADV_* constants are not all defined in
    # Cython's mman.pxd, so we add them in addition to our feature flags.

    # FreeBSD
    enum: NOSYNC
    enum: MADV_NOSYNC

    enum: AUTOSYNC
    enum: MADV_AUTOSYNC

    enum: NOCORE
    enum: MADV_NOCORE

    enum: CORE
    enum: MADV_CORE

    enum: PROTECT
    enum: MADV_PROTECT

    # OpenBSD
    enum: SPACEAVAIL
    enum: MADV_SPACEAVAIL

    # Solaris
    enum: ACCESS_DEFAULT
    enum: ACCESS_LWP
    enum: ACCESS_MANY
    enum: ACCESS_MANY_PSET
    enum: PURGE
    enum: MADV_ACCESS_DEFAULT
    enum: MADV_ACCESS_LWP
    enum: MADV_ACCESS_MANY
    enum: MADV_ACCESS_MANY_PSET
    enum: MADV_PURGE



# Always prefer setuptools over distutils
from setuptools import setup, find_packages
from distutils.extension import Extension
from os import path

try:
    from Cython.Build import cythonize
    USE_CYTHON = True
except ImportError:
    USE_CYTHON = False


ext = '.pyx' if USE_CYTHON else '.c'
extensions = [Extension("fmmap", ["src/fmmap"+ext], include_dirs=["src"])]

if USE_CYTHON:
    from Cython.Build import cythonize
    extensions = cythonize(extensions)


here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='fmmap',
    version='2.0.0',
    description='A fast reimplementation of mmap',
    long_description=long_description,
    author='Friedel Wolff',
    author_email='friedel@translate.org.za',

    # For a list of valid classifiers, see https://pypi.org/classifiers/
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Programming Language :: Cython',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Operating System :: POSIX :: Linux',
        'Operating System :: POSIX :: BSD :: FreeBSD',
        'Operating System :: POSIX :: BSD :: NetBSD',
        'Operating System :: POSIX :: BSD :: OpenBSD',
        'Operating System :: POSIX :: SunOS/Solaris',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
    ],

    # Note that this is a string of words separated by whitespace, not a list.
    keywords='mmap',
    package_dir={'': 'src'},
    ext_modules=extensions,

    # Specify which Python versions you support. In contrast to the
    # 'Programming Language' classifiers above, 'pip install' will check this
    # and refuse to install the project if the version does not match. If you
    # do not support Python 2, you can simplify this to '>=3.5' or similar, see
    # https://packaging.python.org/guides/distributing-packages-using-setuptools/#python-requires
    python_requires='>=3.4, <4',

    # List additional groups of dependencies here (e.g. development
    # dependencies). Users will be able to install these using the "extras"
    # syntax, for example:
    #
    #   $ pip install sampleproject[dev]
    #
    # Similar to `install_requires` above, these must be valid existing
    # projects.
    extras_require={  # Optional
        'dev': ['check-manifest'],
        'test': ['coverage'],
    },

    url='https://github.com/friedelwolff/fmmap/',
    # This field corresponds to the "Project-URL" metadata fields:
    # https://packaging.python.org/specifications/core-metadata/#project-url-multiple-use
    project_urls={
        'Bug Reports': 'https://github.com/friedelwolff/fmmap/issues',
        'Source': 'https://github.com/friedelwolff/fmmap/',
    },
)
